
TidyPlatesUtility:EnableGroupWatcher()
TidyPlatesWidgets:EnableAuraWatcher()
TidyPlatesWidgets:EnableTankWatch()

TidyPlatesHubFunctions = {}
local LocalVars = TidyPlatesHubDamageVariables
local LocalRole = 1

local WidgetLib = TidyPlatesWidgets
local valueToString = TidyPlatesUtility.abbrevNumber
local EnableTankWatch = TidyPlatesWidgets.EnableTankWatch
local DisableTankWatch = TidyPlatesWidgets.DisableTankWatch
local IsTankedByAnotherTank = TidyPlatesWidgets.IsTankedByAnotherTank
local GetThreatCondition = TidyPlatesWidgets.GetThreatCondition
local IsTotem = TidyPlatesUtility.IsTotem
local IsAuraShown = TidyPlatesWidgets.IsAuraShown

local CreateThreatLineWidget = WidgetLib.CreateThreatLineWidget
local CreateAuraWidget = WidgetLib.CreateAuraWidget
local CreateClassWidget = WidgetLib.CreateClassWidget
local CreateRangeWidget = WidgetLib.CreateRangeWidget
local CreateComboPointWidget = WidgetLib.CreateComboPointWidget
local CreateTotemIconWidget = WidgetLib.CreateTotemIconWidget

local BlueColor = {r = 60/255, g =  168/255, b = 255/255, }
local GreenColor = { r = 96/255, g = 224/255, b = 37/255, }
local RedColor = { r = 255/255, g = 51/255, b = 32/255, }
local YellowColor = { r = 252/255, g = 220/255, b = 27/255, }
local GoldColor = { r = 252/255, g = 140/255, b = 0, }
local OrangeColor = { r = 255/255, g = 64/255, b = 0, }
local WhiteColor = { r = 250/255, g = 250/255, b = 250/255, }

local function DummyFunction() end

---------------------------------------
-- Text
---------------------------------------
-- unit.health, unit.healthmax, unit.targetOf

-- None
local function HealthFunctionNone() return "" end
-- Percent
local function HealthFunctionPercent(unit) 
		return "%"..ceil(100*(unit.health/unit.healthmax)) 
end
-- Actual
local function HealthFunctionActual(unit) 
	return valueToString(unit.health) 
end
--Deficit
local function HealthFunctionDeficit(unit) 
	local health, healthmax = unit.health, unit.healthmax
	if health ~= healthmax then return "-"..valueToString(healthmax - health) end 
end
-- Total and Percent
local function HealthFunctionTotal(unit) 
	local health, healthmax = unit.health, unit.healthmax
	return valueToString(health).." / "..valueToString(healthmax).." ("..ceil(100*(health/healthmax)).."%)" 
end
-- TargetOf
local function HealthFunctionTargetOf(unit) 
	if unit.isTarget then return UnitName("targettarget")
	elseif unit.isMouseover then return UnitName("mouseovertarget")
	else return "" end
end

local HealthTextModeFunctions = {
	HealthFunctionNone,
	HealthFunctionPercent,
	HealthFunctionActual,
	HealthFunctionDeficit,
	HealthFunctionTotal,
	HealthFunctionTargetOf,
}

local function HealthTextDelegate(unit) 
	return HealthTextModeFunctions[LocalVars.TextHealthTextMode](unit) 
end

---------------------------------------
-- Scale
---------------------------------------

-- By Elite
local function ScaleFunctionByElite(unit) 
	if unit.isElite then return LocalVars.ScaleSpotlight end 
end

-- By Target
local function ScaleFunctionByTarget(unit) 
	if unit.isTarget then return LocalVars.ScaleSpotlight end 
end

-- By Threat (High) DPS Mode
local function ScaleFunctionByThreatHigh(unit)
	if InCombatLockdown() and unit.reaction == "HOSTILE" then 
		if unit.type == "NPC" and unit.threatValue > 1 and unit.health > 0 then return LocalVars.ScaleSpotlight end
	end
end

-- By Threat (Low) Tank Mode
local function ScaleFunctionByThreatLow(unit)
	if InCombatLockdown() and unit.reaction == "HOSTILE" then 
		if IsTankedByAnotherTank(unit) then return end
		if unit.type == "NPC" and unit.health > 0 and unit.threatValue < 2 then return LocalVars.ScaleSpotlight end 
	end
end

-- By Debuff Widget
local function ScaleFunctionByActiveDebuffs(unit, frame) 
	local widget = unit.frame.widgets.DebuffWidget
	--local widget = TidyPlatesWidgets.GetAuraWidgetByGUID(unit.guid)
	if IsAuraShown(widget) then return LocalVars.ScaleSpotlight end
end

-- By Enemy
local function ScaleFunctionByEnemy(unit) 
	if unit.reaction ~= "FRIENDLY" then return LocalVars.ScaleSpotlight end 
end

-- By NPC
local function ScaleFunctionByNPC(unit) 
	if unit.type == "NPC" then return LocalVars.ScaleSpotlight end 
end

local ScaleFunctionsDamage = { DummyFunction, ScaleFunctionByElite, ScaleFunctionByTarget, ScaleFunctionByThreatHigh, ScaleFunctionByActiveDebuffs, ScaleFunctionByEnemy,ScaleFunctionByNPC}
local ScaleFunctionsTank = { DummyFunction, ScaleFunctionByElite, ScaleFunctionByTarget, ScaleFunctionByThreatLow, ScaleFunctionByActiveDebuffs, ScaleFunctionByEnemy, ScaleFunctionByNPC}

-- Scale Functions Listed by Role order: Damage, Tank, Heal
local ScaleFunctions = {ScaleFunctionsDamage, ScaleFunctionsTank}

local function ScaleDelegate(...)
	local unit = ...
	local scale
	
	if LocalVars.ScaleIgnoreNonEliteUnits and (not unit.isElite) then
	elseif LocalVars.ScaleIgnoreNeutralUnits and unit.reaction == "NEUTRAL" then 
	else scale = ScaleFunctions[LocalRole][LocalVars.ScaleSpotlightMode](...)  end
	return scale or LocalVars.ScaleStandard
end


---------------------------------------
-- Opacity / Alpha
---------------------------------------

-- By Threat (High)
local function AlphaFunctionByThreatHigh (unit) 
	if InCombatLockdown() and unit.reaction == "HOSTILE" then 
		if unit.threatValue > 1 and unit.health > 0 then return LocalVars.OpacitySpotlight end
	end
end

-- Tank Mode
local function AlphaFunctionByThreatLow (unit) 
	if InCombatLockdown() and unit.reaction == "HOSTILE" then 
		if IsTankedByAnotherTank(unit) then return end
		if unit.threatValue < 2 and unit.health > 0 then return LocalVars.OpacitySpotlight end
	end
end

local function AlphaFunctionByMouseover(unit) 
	if unit.isMouseover then return LocalVars.OpacitySpotlight end
end

local function AlphaFunctionByEnemy(unit) 
	if unit.reaction ~= "FRIENDLY" then return LocalVars.OpacitySpotlight end
end

local function AlphaFunctionByNPC(unit) 
	if unit.type == "NPC" then return LocalVars.OpacitySpotlight end 
end

local function AlphaFunctionByActiveDebuffs(unit) 
	local widget = unit.frame.widgets.DebuffWidget
	--local widget = TidyPlatesWidgets.GetAuraWidgetByGUID(unit.guid)
	if IsAuraShown(widget) then return LocalVars.OpacitySpotlight end
end

local function AlphaFilter(unit)
	if LocalVars.OpacityFilterCustom[unit.name] then return true
	elseif LocalVars.OpacityFilterNeutralUnits and unit.reaction == "NEUTRAL" then return true
	elseif LocalVars.OpacityFilterNonElite and (not unit.isElite) then return true
	end
end

local AlphaFunctionsDamage = { DummyFunction, AlphaFunctionByThreatHigh, AlphaFunctionByMouseover, AlphaFunctionByActiveDebuffs, AlphaFunctionByEnemy, AlphaFunctionByNPC, }
local AlphaFunctionsTank = { DummyFunction, AlphaFunctionByThreatLow, AlphaFunctionByMouseover, AlphaFunctionByActiveDebuffs, AlphaFunctionByEnemy, AlphaFunctionByNPC, }

-- Alpha Functions Listed by Role order: Damage, Tank, Heal
local AlphaFunctions = {AlphaFunctionsDamage, AlphaFunctionsTank}

local function AlphaDelegate(...)
	local unit = ...
	local alpha
	
	if unit.isTarget then return LocalVars.OpacityTarget
	elseif unit.isCasting and LocalVars.OpacityFullSpell then return LocalVars.OpacityTarget
	else
		-- Filter
		if AlphaFilter(unit) then alpha = LocalVars.OpacityFiltered 
		-- Spotlight
		else alpha = AlphaFunctions[LocalRole][LocalVars.OpacitySpotlightMode](...) end
	end
	
	if alpha then return alpha
	else 
		if (not UnitExists("target")) and LocalVars.OpacityFullNoTarget then return LocalVars.OpacityTarget
		else return LocalVars.OpacityNonTarget end
	end
end


---------------------------------------
-- Health Bar Color
---------------------------------------

--"By Class"
local function ColorFunctionByClass(unit)
	--
end


local function ColorFunctionsDamage(unit)
	if unit.threatValue > 1 then return LocalVars.ColorAttackingMe
	elseif unit.threatValue == 1 then return LocalVars.ColorAggroTransition
	else return LocalVars.ColorAttackingOthers end
end

local function ColorFunctionsTank(unit)
	if unit.threatValue > 2 then return LocalVars.ColorAttackingMe
	elseif unit.threatValue == 2 then return LocalVars.ColorAggroTransition
	else
		if IsTankedByAnotherTank(unit) then return LocalVars.ColorAttackingOtherTank end
		return LocalVars.ColorAttackingOthers 
	end
end

-- By Threat Color Functions Listed by Role order: Damage, Tank, Heal
local ByThreatColorRoleFunctions = {ColorFunctionsDamage, ColorFunctionsTank}

--"By Aggro"
local function ColorFunctionByThreat(unit)
	if InCombatLockdown() and unit.reaction ~= "FRIENDLY" and unit.type == "NPC" then
		--if IsTankedByAnotherTank(unit) then return LocalVars.ColorAttackingOtherTank end
		return ByThreatColorRoleFunctions[LocalRole](unit)
	end
end

--"By Reaction"
local ReactionColors = {
	["FRIENDLY"] = {
		["PLAYER"] = {r = 0, g = 0, b = 1,},
		["NPC"] = {r = 0, g = 1, b = 0,},
	},
	["HOSTILE"] = {
		["PLAYER"] = {r = 1, g = 0, b = 0,},
		["NPC"] = {r = 1, g = 0, b = 0,},
	},
	["NEUTRAL"] = {
		["NPC"] = {r = 1, g = 1, b = 0,},
	},
}

local function ColorFunctionByReaction(unit)
	return ReactionColors[unit.reaction][unit.type]
end
	
local ColorFunctions = {DummyFunction, ColorFunctionByClass, ColorFunctionByThreat, ColorFunctionByReaction }

local function HealthColorDelegate(unit)
	local color	
	
	if LocalVars.ClassColorPartyMembers and unit.reaction == "FRIENDLY" and unit.type == "PLAYER"  then
		local class = TidyPlatesUtility.GroupMembers.Class[unit.name]
		if class then color = RAID_CLASS_COLORS[class] end
	else
		color = ColorFunctions[LocalVars.ColorHealthBarMode](unit)
	end
	
	if color then return color.r, color.g, color.b 
	else return unit.red, unit.green, unit.blue end
end

---------------------------------------
-- Warning Border Color
---------------------------------------

-- "By Threat (High) Damage"
local function WarningBorderFunctionByThreatDamage(unit)
	if InCombatLockdown() and unit.reaction ~= "FRIENDLY" and unit.type == "NPC" then
		if unit.threatValue > 0 then 
			return ColorFunctionsDamage(unit)
		end
	end
end

-- "By Threat (Low) Tank"
local function WarningBorderFunctionByThreatTank(unit)
	if InCombatLockdown() and unit.reaction ~= "FRIENDLY" and unit.type == "NPC" then
		if unit.threatValue < 3 and not IsTankedByAnotherTank(unit) then 
			return ColorFunctionsTank(unit)
		end
	end
end
	
local WarningBorderFunctionsDamage = { DummyFunction, WarningBorderFunctionByThreatDamage }
local WarningBorderFunctionsTank = { DummyFunction, WarningBorderFunctionByThreatTank }
local WarningBorderFunctions = {WarningBorderFunctionsDamage, WarningBorderFunctionsTank}

local function ThreatColorDelegate(unit)
	local color
	
	if LocalVars.ColorDangerGlowOnParty and unit.reaction == "FRIENDLY" then
		if GetThreatCondition(unit.name) then return 1, 0, 0, 1 end
	end

	color = WarningBorderFunctions[LocalRole][LocalVars.ColorDangerGlowMode](unit)
	if color then return color.r, color.g, color.b, 1
	else return 0, 0, 0, 0 end
end


---------------------------------------
-- Cast Bar Color
---------------------------------------
local function CastBarDelegate(unit)
	local color
	if unit.spellIsShielded then color = WhiteColor -- OrangeColor or WhiteColor
	else color = GoldColor end 
	return color.r, color.g, color.b, 1
end


---------------------------------------
-- Widgets
---------------------------------------
--[[
	CurrentDebuff.name = GetSpellInfo(spellid)
	CurrentDebuff.spellid = spellid
	CurrentDebuff.stacks = stacks
	CurrentDebuff.duration = duration
	CurrentDebuff.caster = caster
	--UnitGUID("player")
--]]
local DebuffFilterModes = {
	-- All
	function(debuff) 
		return true
	end,
	-- Specific	
	function(debuff) 
		if LocalVars.WidgetsDebuffList[debuff.name] then return true end
	end,
	-- All mine	
	function(debuff) 
		if debuff.caster == UnitGUID("player") then return true end
	end,
	-- My specific	
	function(debuff) 
		if LocalVars.WidgetsDebuffList[debuff.name] and debuff.caster == UnitGUID("player") then return true end
	end,
}

local function DebuffFilter(debuff)
	return DebuffFilterModes[LocalVars.WidgetsDebuffMode](debuff)
end

	
local function AddClassIcon(plate, enable, config)
	if enable then 
		if not plate.widgets.ClassIcon then
			local widget
			widget = CreateClassWidget(plate)
			widget:SetPoint(config.anchor or "TOP", plate, config.x or 0, config.y or 0) -- 0, 3)
			--widget:SetScale(1.2)
			plate.widgets.ClassIcon = widget
		end
	elseif plate.widgets.ClassIcon then 
		plate.widgets.ClassIcon:Hide()
		plate.widgets.ClassIcon = nil
	end
end

local function AddTotemIcon(plate, enable, config)
	if enable then 
		if not plate.widgets.TotemIcon then
			local widget
			widget = CreateTotemIconWidget(plate)
			widget:SetPoint(config.anchor or "TOP", plate, config.x or 0, config.y or 0) --0, 3)
			--widget:SetScale(1.2)
			plate.widgets.TotemIcon = widget
		end
	elseif plate.widgets.TotemIcon then 
		plate.widgets.TotemIcon:Hide()
		plate.widgets.TotemIcon = nil
	end
end

local function AddComboPoints(plate, enable, config)
	if enable then 
		if not plate.widgets.ComboWidget then
			local widget
			widget = CreateComboPointWidget(plate)
			widget:SetPoint(config.anchor or "TOP", plate, config.x or 0, config.y or 0) --0, 10)
			widget:SetFrameLevel(plate:GetFrameLevel()+2)
			plate.widgets.ComboWidget = widget
		end
	elseif plate.widgets.ComboWidget then 
		plate.widgets.ComboWidget:Hide()
		plate.widgets.ComboWidget = nil
	end
end

local function AddThreatWidget(plate, enable, config)
	if enable then 
		if not plate.widgets.ThreatWidget then
			local widget
			widget = CreateThreatLineWidget(plate)
			widget:SetPoint(config.anchor or "TOP", plate, config.x or 0, config.y or 0) --0, 4)
			widget:SetFrameLevel(plate:GetFrameLevel()+3)
			--widget._LowColor = LocalVars.TugWidgetLooseColor
			--widget._HighColor = LocalVars.TugWidgetAggroColor
			--widget._TankedColor = LocalVars.TugWidgetSafeColor
			plate.widgets.ThreatWidget = widget
		end
	elseif plate.widgets.ThreatWidget then 
		plate.widgets.ThreatWidget:Hide()
		plate.widgets.ThreatWidget = nil
	end
end

local RangeModeRef = { 9, 15, 28, 40 }
local function AddRangeWidget(plate, enable, config)
	if enable then 
		if not plate.widgets.RangeWidget then
			local widget
			widget = CreateRangeWidget(plate)
			widget:SetPoint(config.anchor or "CENTER", config.x or 0, config.y or 0) --0, 0)
			plate.widgets.RangeWidget = widget
		end
	elseif plate.widgets.RangeWidget then 
		plate.widgets.RangeWidget:Hide()
		plate.widgets.RangeWidget = nil 
	end
end

local function AddDebuffWidget(plate, enable, config)
	if enable then 
		if not plate.widgets.DebuffWidget then
			local widget
			widget =  CreateAuraWidget(plate)
			widget:SetPoint(config.anchor or "TOP", plate, config.x or 0, config.y or 0) --15, 20)
			widget:SetFrameLevel(plate:GetFrameLevel()+1)
			widget.Filter = DebuffFilter
			plate.widgets.DebuffWidget = widget
		end
	elseif plate.widgets.DebuffWidget then 
		plate.widgets.DebuffWidget:Hide()
		plate.widgets.DebuffWidget = nil
	 end
end


---------------------------------------
-- Widget Activation
---------------------------------------

local function OnInitializeWidgets(plate, configTable)
	AddClassIcon(plate, ((LocalVars.ClassEnemyIcon or LocalVars.ClassPartyIcon)) , configTable.ClassIcon)
	AddTotemIcon(plate,  LocalVars.WidgetsTotemIcon, configTable.TotemIcon)
	AddThreatWidget(plate, LocalVars.WidgetsThreatIndicator, configTable.ThreatWidget, LocalVars.WidgetsThreatIndicatorMode )
	AddComboPoints(plate, LocalVars.WidgetsComboPoints, configTable.ComboWidget )
	AddRangeWidget(plate, LocalVars.WidgetsRangeIndicator, configTable.RangeWidget )
	AddDebuffWidget(plate, LocalVars.WidgetsDebuff, configTable.DebuffWidget )
end



local function OnContextUpdateDelegate(plate, unit)
	local Widgets = plate.widgets
	if LocalVars.WidgetsComboPoints then Widgets.ComboWidget:UpdateContext(unit) end
	if LocalVars.WidgetsThreatIndicator then Widgets.ThreatWidget:UpdateContext(unit) end
	if LocalVars.WidgetsDebuff then Widgets.DebuffWidget:UpdateContext(unit) end
end

local function OnUpdateDelegate(plate, unit)
	local Widgets = plate.widgets
	if LocalVars.WidgetsRangeIndicator then Widgets.RangeWidget:Update(unit,RangeModeRef[LocalVars.RangeMode])  end
	if LocalVars.ClassEnemyIcon or LocalVars.ClassPartyIcon then Widgets.ClassIcon:Update(unit, LocalVars.ClassPartyIcon)  end
	if LocalVars.WidgetsTotemIcon then Widgets.TotemIcon:Update(unit)  end
end

local function UseDamageVariables()
	SetCVar("threatWarning", 3)		-- Required for threat/aggro detection
	LocalVars = TidyPlatesHubDamageVariables
	LocalRole = 1
	return TidyPlatesHubDamageVariables
end

local function UseTankVariables()
	SetCVar("threatWarning", 3)		-- Required for threat/aggro detection
	LocalVars = TidyPlatesHubTankVariables
	LocalRole = 2
	return TidyPlatesHubTankVariables
end

---------------
-- Function List
---------------
TidyPlatesHubFunctions.SetScale = ScaleDelegate
TidyPlatesHubFunctions.SetAlpha = AlphaDelegate
TidyPlatesHubFunctions.SetCustomText = HealthTextDelegate
TidyPlatesHubFunctions.OnUpdate = OnUpdateDelegate
TidyPlatesHubFunctions.OnInitializeWidgets = OnInitializeWidgets
TidyPlatesHubFunctions.SetHealthbarColor = HealthColorDelegate
TidyPlatesHubFunctions.SetThreatColor = ThreatColorDelegate
TidyPlatesHubFunctions.OnContextUpdate = OnContextUpdateDelegate
TidyPlatesHubFunctions.SetCastbarColor = CastBarDelegate
TidyPlatesHubFunctions.UseDamageVariables = UseDamageVariables
TidyPlatesHubFunctions.UseTankVariables = UseTankVariables

